/*
    Ming, an SWF output library
    Copyright (C) 2001  Opaque Industries - http://www.opaque.net/

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "_cl.h"

#include <math.h>

#include "font.h"
#include "method.h"

#define glyphLength(font,glyph) \
  ((font)->glyphOffset[(glyph)+1] - (font)->glyphOffset[(glyph)])

int completeSWFFont(SWFBlock block)
{
  SWFFont font = (SWFFont)block;
  int size, i;

  SWFFont_resolveTextList(font);

  /* 2 bytes id, 2 bytes flags, 1 byte name length, 2 bytes nGlyphs,
     2*(nGlyphs+1) bytes offsets, nGlyphs bytes character map: */

  size = 9 + strlen(font->name) + 3*font->nGlyphs;

  /* get length of each glyph from its output buffer */
  for(i=0; i<font->nGlyphs; ++i)
    size += glyphLength(font, font->codeToGlyph[i]);

  if(size > 65500)
  {
    size += 2*(font->nGlyphs+1);
    font->flags |= SWF_FONT_WIDEOFFSETS;
  }

  return size;
}
void writeSWFFontToMethod(SWFBlock block,
			  SWFByteOutputMethod method, void *data)
{
  SWFFont font = (SWFFont)block;
  int offset, i;
  byte *p, *s;

  methodWriteUInt16(CHARACTERID(font), method, data);

  method((byte)(font->flags & SWF_FONT_WIDEOFFSETS ? 8 : 0), data); /* main flags */
  method(0, data);                                          /* more flags */
  method((byte)(strlen(font->name)), data);

  for(p = font->name; *p != '\0'; ++p)
    method(*p, data);

  methodWriteUInt16(font->nGlyphs, method, data);

  offset = (font->nGlyphs+1)*(font->flags & SWF_FONT_WIDEOFFSETS ? 4 : 2);

  /* write offset table for glyphs */
  for(i=0; i<=font->nGlyphs; ++i)
  {
    if(font->flags & SWF_FONT_WIDEOFFSETS)
      methodWriteUInt32(offset, method, data);
    else
      methodWriteUInt16(offset, method, data);

    if(i < font->nGlyphs)
      offset += glyphLength(font, font->codeToGlyph[i]);
  }

  /* write shape records for glyphs */
  for(i=0; i<font->nGlyphs; ++i)
  {
    p = font->glyphOffset[font->codeToGlyph[i]];
    s = font->glyphOffset[font->codeToGlyph[i]+1];

    SWF_assert(p < s);

    while(p < s)
      method(*(p++), data);
  }

  /* write (dummy) character mapping */
  for(i=0; i<font->nGlyphs; ++i)    /* this is rubbish - would need another */
    method((byte)(i), data);                /* table to reverse mapping */
}

void destroySWFFont(SWFBlock block)
{
  SWFFont font = (SWFFont)block;
  MingUnRecordNewDestructibleBlockRemoved((SWFBlock)font);

  if(font->shapes)
    _free_(font->shapes);
  if(font->bounds)
    _free_(font->bounds);
  if(font->name)
    _free_(font->name);
  if(font->kernTable)
    _free_(font->kernTable);

  _free_(font);
}
WM_DLL_A SWFFont WM_DLL_B newSWFFont()
{
  SWFFont font = _calloc_(1, SWFFONT_SIZE);

  CHARACTER(font)->number = ++SWF_gNumCharacters;
  BLOCK(font)->type = SWF_DEFINEFONT2;
  BLOCK(font)->writeBlock = writeSWFFontToMethod;
  BLOCK(font)->complete = completeSWFFont;
  BLOCK(font)->dtor = destroySWFFont;

  memset(&(font->glyphToCode), 0xff, 256);
  MingRecordNewDestructibleBlockAdded((SWFBlock)(font));

  return font;
}
 
WM_DLL_A byte * WM_DLL_B SWFFont_findCharacterGlyph(SWFFont font, unsigned char c)
{
  return font->glyphOffset[font->codeTable[c]];
}

WM_DLL_A void WM_DLL_B SWFFont_addTextToList(SWFFont font, struct _textRecord *text)
{
  SWFTextList textList = _calloc_(1, TEXTLIST_SIZE);
  textList->next = NULL;
  textList->text = text;

  if(font->currentList != NULL)
    font->currentList->next = textList;
  else
    font->textList = textList;

  font->currentList = textList;
}

WM_DLL_A void WM_DLL_B SWFFont_addCharToTable(SWFFont font, byte c)
{
  if(font->glyphToCode[c]==0xff) /* assuming one can't actually use all 255 */
  {
    font->codeToGlyph[font->nGlyphs] = font->codeTable[c];
    font->glyphToCode[c] = font->nGlyphs;
    ++(font->nGlyphs);
  }
}

/* XXX - big confusion between terms here.  CodeTable isn't font->codeTable */
WM_DLL_A void WM_DLL_B SWFFont_buildCodeTable(SWFFont font, SWFTextRecord text)
{
  SWFTextRecord textRecord;
  const byte *string;
  int l, i;

  textRecord = text;
  while(textRecord != NULL)
  {
    string = textRecord->string;

    if(string != NULL)
    {
      l = strlen(string);
      for(i=0; i<l; ++i)
      	SWFFont_addCharToTable(font, string[i]);
    }

    textRecord = textRecord->next;
  }
}

/* build code table from text in all proceding Text blocks */
WM_DLL_A void WM_DLL_B SWFFont_resolveTextList(SWFFont font)
{
  SWFTextList textList, oldList;

  textList = font->textList;
  while(textList != NULL)
  {
    oldList = textList;
    SWFFont_buildCodeTable(font, textList->text);
    textList = textList->next;
    _free_(oldList);
  }

  font->textList = NULL;
}

/* return length of given string in whatever units these are we're using */
WM_DLL_A int WM_DLL_B SWFFont_getScaledStringWidth(SWFFont font, const unsigned char *string)
{
  int i, j, width = 0, l = strlen(string);
  int glyph, glyph2;

  for(i=0; i<l; ++i)
  {
    glyph = font->codeTable[string[i]];

    if(font->advances)
      width += font->advances[glyph];

    /* looking in kernTable */
    if (i<l-1 && font->kernTable)
    {
      glyph2 = font->codeTable[string[i+1]];

      j=font->kernCount;

      while (--j >= 0)
      {
	if (glyph == font->kernTable[j].code1 &&
	    glyph2 == font->kernTable[j].code2)
	{
	  width += font->kernTable[j].adjustment;
	  break;
	}
      }
    }
  }

  return width;
}

/* get some font metrics */
WM_DLL_A short WM_DLL_B SWFFont_getScaledAscent(SWFFont font)
{
  return font->ascent;
}
WM_DLL_A short WM_DLL_B SWFFont_getScaledDescent(SWFFont font)
{
  return font->descent;
}
WM_DLL_A short WM_DLL_B SWFFont_getScaledLeading(SWFFont font)
{
  return font->leading;
}

     
                                   